import json
import pathlib

import pytest
from pyvfg import (
    VFG,
    DiscreteVariableAnonymousElements,
    DiscreteVariableNamedElements,
    Distribution,
    Factor,
    FactorRole,
    IncorrectProbabilityLength,
    InvalidVariableName,
    ValidationError,
    Variable,
    VariableRole,
    validate,
    vfg_from_json,
    vfg_to_json_schema,
)

from .. import package_dir


def test_vfg_to_dict_default():
    vfg = VFG()
    vfg_dict = vfg.to_dict()
    assert vfg_dict["variables"] == {}
    assert vfg_dict["factors"] == []


def test_vfg_to_dict_custom(variables_1, factors_1, metadata_1, vis_1):
    vfg = VFG(variables=variables_1, factors=factors_1)
    vfg.metadata = metadata_1
    vfg.visualization_metadata = vis_1
    vfg_dict = vfg.to_dict()

    assert vfg_dict["variables"] == variables_1
    assert vfg_dict["factors"][0]["variables"] == factors_1[0]["variables"]
    assert vfg_dict["factors"][0]["distribution"] == factors_1[0]["distribution"]
    assert vfg_dict["factors"][0]["values"] == factors_1[0]["values"]
    assert vfg_dict["factors"][0]["role"] == factors_1[0]["role"]
    assert vfg_dict["metadata"] == metadata_1
    assert vfg_dict["visualization_metadata"] == vis_1


def test_vfg_to_json_schema():
    vfg_schema_dict, _ = vfg_to_json_schema()

    assert vfg_schema_dict["title"] == "VFG"
    assert vfg_schema_dict["type"] == "object"
    assert vfg_schema_dict["properties"]["variables"]["type"] == "object"
    assert vfg_schema_dict["properties"]["version"]["default"] == "0.4.0"
    assert vfg_schema_dict["properties"]["factors"]["type"] == "array"


# Valid VFG examples
@pytest.fixture
def valid_vfg_1():
    with open(
        package_dir / "tests/fixtures/models/sprinkler/sprinkler_vfg_0_3_0.json", "r"
    ) as f:
        vfg_json = json.load(f)
    return VFG(**vfg_json)


@pytest.fixture
def valid_vfg_2():
    with open(
        package_dir / "tests/fixtures/models/insurance/insurance_vfg_0_4_0.json", "r"
    ) as f:
        vfg_json = json.load(f)
    return VFG(**vfg_json)


# Invalid VFG examples
@pytest.fixture
def invalid_vfg_1():
    variables = {
        "x?": Variable(DiscreteVariableNamedElements(elements=["a", "b", "c"])),
        "y": Variable(DiscreteVariableAnonymousElements(cardinality=2)),
    }
    factors = [
        Factor(
            variables=["x?", "y"],
            distribution=Distribution.Categorical,
            values=[0.1, 0.2, 0.3],
        )
    ]
    return VFG(variables=variables, factors=factors)


@pytest.fixture
def invalid_vfg_2():
    variables = {
        "a": Variable(DiscreteVariableNamedElements(elements=["x", "y", "z"])),
        "b": Variable(DiscreteVariableAnonymousElements(cardinality=2)),
    }
    factors = [
        Factor(
            variables=["a", "b"],
            distribution=Distribution.Categorical,
            values=[0.1, 0.2, 0.3],
        )
    ]
    return VFG(variables=variables, factors=factors)


def test_validate_valid_vfg(valid_vfg_1, valid_vfg_2):
    assert validate(valid_vfg_1) is None
    assert validate(valid_vfg_2) is None


def test_validate_invalid_vfg(invalid_vfg_1, invalid_vfg_2):
    try:
        validate(invalid_vfg_1)
        assert False, "should not reach here"
    except InvalidVariableName as e:
        assert "x?" in str(e)

    try:
        validate(invalid_vfg_2)
        assert False, "should not reach here"
    except IncorrectProbabilityLength as e:
        assert "incorrect number of probability values. Expected 1, found 2" in str(e)


def test_vfg_from_json(valid_vfg_1: VFG):
    vfg = vfg_from_json(valid_vfg_1.model_dump())
    assert validate(vfg) is None, "check for validity"
    assert vfg == valid_vfg_1, "check for round trip"
    vfg = vfg_from_json(valid_vfg_1.model_dump_json())
    assert validate(vfg) is None, "check for string validity"
    assert vfg == valid_vfg_1, "check for string round-trip"


@pytest.mark.parametrize(
    "f",
    [
        "post_validate_empty_values_array.json",
        "post_validate_missing_required_fields_factors.json",
    ],
)
def test_validate_graph_failure(f):
    """
    Tests graph validation, with the same procedure as gpil-pipeline.
    :return: None
    """
    import os

    import pyvfg

    basedir = os.path.join(os.path.dirname(__file__), "..", "payloads", "validate")
    with open(os.path.join(basedir, f), "r") as file:
        vfg_json = file.read()
        vfg = pyvfg.vfg_upgrade(vfg_json)
        with pytest.raises(pyvfg.ValidationError):
            pyvfg.validate_graph(vfg)


@pytest.mark.parametrize("f", ["post_validate_example.json"])
def test_validate_graph_success(f):
    """
    Tests graph validation, with the same procedure as gpil-pipeline.
    :return:
    """
    import os

    import pyvfg

    basedir = os.path.join(os.path.dirname(__file__), "..", "payloads", "validate")
    with open(os.path.join(basedir, f), "r") as file:
        vfg_json = file.read()
        vfg = pyvfg.vfg_upgrade(vfg_json)
        # NOTE: Change from the above! we want these to succeed at validation
        assert pyvfg.validate_graph(vfg) is None


if __name__ == "__main__":
    vfg_schema_dict, vfg_schema_json = vfg_to_json_schema()
    print(json.dumps(vfg_schema_dict, indent=2))
